"""Albumart tags

Abstraction for album art image format processing

"""

import os
import requests
import StringIO

from PIL import ImageFile

DEFAULT_ARTWORK_FILENAME = 'artwork.jpg'

PIL_EXTENSION_MAP = {
    'JPEG':     'jpg',
    'PNG':      'png',
}

PIL_MIME_MAP = {
    'JPEG':     'image/jpeg',
    'PNG':      'image/png',
}

class AlbumArtError(Exception):
    pass

class AlbumArt(object):
    """
    Class to parse albumart image files from tags and files
    """

    def __init__(self, path=None):
        self.__image = None
        self.__mimetype = None

        if path is not None:
            self.import_file(path)

    def __repr__(self):
        """
        Returns text description of image type and size
        """

        if not self.is_loaded():
            return 'Uninitialized AlbumArt object.'
        return '%(mime)s %(bytes)d bytes %(width)dx%(height)d' % self.get_info()

    def __unicode__(self):
        """
        Returns file format and size as unicode string
        """

        if not self.is_loaded():
            return unicode('Uninitialized AlbumArt object')
        return unicode('%(mime)s %(width)dx%(height)dpx' % self.get_info())

    def __len__(self):
        """
        Returns PIL image length as string
        """

        if not self.is_loaded():
            return 0
        return len(self.__image.tostring())

    def __parse_image(self, data):
        """
        Load the image from data with PIL
        """

        try:
            parser = ImageFile.Parser()
            parser.feed(data)
            self.__image = parser.close()
        except IOError:
            raise AlbumArtError('Error parsing albumart image data')

        try:
            self.__mimetype = PIL_MIME_MAP[self.__image.format]
        except KeyError:
            self.__image = None
            raise AlbumArtError('Unsupported PIL image format: %s' % self.__image.format )

        if self.__image.mode != 'RGB':
            self.__image = self.__image.convert('RGB')

    def import_data(self, data):
        """
        Import albumart from metadata tag or database as bytes
        """
        self.__parse_image(data)

    def import_file(self, path):
        """
        Import albumart from file
        """
        if not os.path.isfile(path):
            raise AlbumArtError('No such file: %s' % path)
        if not os.access(path, os.R_OK):
            raise AlbumArtError('No permissions to read file: %s' % path)

        self.__parse_image(open(path, 'r').read())

    def is_loaded(self):
        """
        Boolean test to see if album art image is loaded
        """
        return self.__image != None

    def get_fileformat(self):
        """
        Return file format of loaded album art image
        """
        if not self.is_loaded():
            raise AlbumArtError('AlbumArt not yet initialized.')
        return self.__image.format

    def get_info(self):
        """
        Return details of loaded album art image
        """
        if not self.is_loaded():
            raise AlbumArtError('AlbumArt not yet initialized.')
        colors = self.__image.getcolors()
        if colors is None:
            colors = 0
        return {
            'type': 3, # Album cover
            'mime': self.__mimetype,
            'bytes': len(self),
            'width': int(self.__image.size[0]),
            'height': int(self.__image.size[1]),
            'colors': colors,
        }

    def dump(self):
        """
        Returns bytes from the image with StringIO.StringIO read() call
        """
        if not self.is_loaded():
            raise AlbumArtError('AlbumArt not yet initialized.')

        s = StringIO.StringIO()
        self.__image.save(s, self.get_fileformat())
        s.seek(0)
        return s.read()

    def save(self, path, fileformat=None):
        """
        Saves the image data to given target file.

        If target filename exists, it is removed before saving.
        """
        if not self.is_loaded():
            raise AlbumArtError('AlbumArt not yet initialized.')

        if fileformat is None:
            fileformat = self.get_fileformat()

        if os.path.isdir(path):
            path = os.path.join(path, DEFAULT_ARTWORK_FILENAME)

        if os.path.isfile(path):
            try:
                os.unlink(path)
            except IOError, (ecode, emsg):
                raise AlbumArtError('Error removing existing file %s: %s' % (path, emsg) )

        try:
            self.__image.save(path, fileformat)
        except IOError, emsg:
            raise AlbumArtError('Error saving %s: %s' % (path, emsg))

    def fetch(self, url):
        res = requests.get(url)
        if res.status_code!=200:
            raise AlbumArtError('Error fetching url %s (returns %s' % (url, res.status_code))

        if 'content-type' not in res.headers:
            raise AlbumArtError('Response did not include content type header')

        try:
            content_type = res.headers['content-type']
            (prefix, extension) = content_type.split('/', 1)
            if prefix!='image':
                raise AlbumArtError('Content type of data is not supported: %s' % content_type)

        except ValueError:
            raise AlbumArtError('Error parsing content type %s' % res.headers['content-type'])

        return self.import_data(res.content)



# coding=utf-8
"""Soundforest database models

SQLAlchemy models for soundforest configuration and music tree databases

"""

import os
import hashlib
import base64
import json
import pytz
from datetime import datetime

from sqlite3 import Connection as SQLite3Connection
from sqlalchemy import create_engine, event
from sqlalchemy import Column, ForeignKey, Integer, Boolean, String, Date
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker, relationship, backref
from sqlalchemy.types import TypeDecorator, Unicode
from sqlalchemy.ext.declarative import declarative_base

from soundforest import SoundforestError, SOUNDFOREST_USER_DIR
from soundforest.log import SoundforestLogger

logger = SoundforestLogger().default_stream

DEFAULT_DATABASE = os.path.join(SOUNDFOREST_USER_DIR, 'soundforest.sqlite')
Base = declarative_base()


class SafeUnicode(TypeDecorator):

    """SafeUnicode columns

    Safely coerce Python bytestrings to Unicode before passing off to the database.

    """

    impl = Unicode

    def process_bind_param(self, value, dialect):
        if isinstance(value, str):
            value = value.decode('utf-8')
        return value


class Base64Field(TypeDecorator):

    """Base64Field

    Column encoded as base64 to a string field in database

    """

    impl = String

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        return base64.encode(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        return base64.decode(value)


class BasePathNamedModel(object):
    """Base name comparable with name string"""

    def __cmp__(self, other):
        if isinstance(other, basestring):
            return cmp(self.path, other)
        return cmp(self, other)

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __lte__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __gte__(self, other):
        return self.__cmp__(other) >= 0


class BaseNamedModel(object):
    """Base name comparable with name string"""

    def __cmp__(self, other):
        if isinstance(other, basestring):
            return cmp(self.name, other)
        return cmp(self, other)

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __lte__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __gte__(self, other):
        return self.__cmp__(other) >= 0


class SettingModel(Base):

    """SettingModel

    Soundforest internal application preferences

    """

    __tablename__ = 'settings'

    id = Column(Integer, primary_key=True)
    key = Column(SafeUnicode, unique=True)
    value = Column(SafeUnicode)


class SyncTargetModel(Base, BaseNamedModel):

    """SyncTargetModel

    Library tree synchronization target entry

    """

    __tablename__ = 'sync_targets'

    id = Column(Integer, primary_key=True)
    name = Column(SafeUnicode, unique=True)
    type = Column(SafeUnicode)
    src = Column(SafeUnicode)
    dst = Column(SafeUnicode)
    flags = Column(SafeUnicode)
    defaults = Column(Boolean)

    def __repr__(self):
        return '%s %s from %s to %s (flags %s)' % (
            self.name, self.type, self.src, self.dst, self.flags
        )

    def as_dict(self):
        return {
            'name': self.name,
            'type': self.type,
            'src':  self.src,
            'dst':  self.dst,
            'flags': self.flags,
            'defaults': self.defaults,
        }


class CodecModel(Base, BaseNamedModel):

    """CodecModel

    Audio format codecs

    """

    __tablename__ = 'codecs'

    id = Column(Integer, primary_key=True)
    name = Column(SafeUnicode)
    description = Column(SafeUnicode)

    def __repr__(self):
        return self.name

    def register_extension(self, session, extension):
        existing = session.query(ExtensionModel).filter(
            ExtensionModel.extension == extension
        ).first()
        if existing:
            raise SoundforestError('ExtensionModel already registered: %s' % extension)

        session.add(ExtensionModel(codec=self, extension=extension))
        session.commit()

    def unregister_extension(self, session, extension):
        existing = session.query(ExtensionModel).filter(
            ExtensionModel.extension == extension
        ).first()
        if not existing:
            raise SoundforestError('ExtensionModel was not registered: %s' % extension)

        session.delete(existing)
        session.commit()

    def register_decoder(self, session, command):
        existing = session.query(DecoderModel).filter(
            DecoderModel.codec == self,
            DecoderModel.command == command
        ).first()
        if existing:
            raise SoundforestError('DecoderModel already registered: %s' % command)

        session.add(DecoderModel(codec=self, command=command))
        session.commit()

    def unregister_decoder(self, session, command):
        existing = session.query(DecoderModel).filter(
            DecoderModel.codec == self,
            DecoderModel.command == command
        ).first()
        if not existing:
            raise SoundforestError('DecoderModel was not registered: %s' % command)

        session.delete(existing)
        session.commit()

    def register_encoder(self, session, command):
        existing = session.query(EncoderModel).filter(
            EncoderModel.codec == self,
            EncoderModel.command == command
        ).first()
        if existing:
            raise SoundforestError('EncoderModel already registered: %s' % command)

        session.add(EncoderModel(codec=self, command=command))
        session.commit()

    def unregister_encoder(self, session, command):
        existing = session.query(EncoderModel).filter(
            EncoderModel.codec == self,
            EncoderModel.command == command
        ).first()
        if not existing:
            raise SoundforestError('EncoderModel was not registered: %s' % command)

        session.delete(existing)
        session.commit()

    def register_tester(self, session, command):
        existing = session.query(TesterModel).filter(
            TesterModel.codec == self,
            TesterModel.command == command
        ).first()
        if existing:
            raise SoundforestError('File tester already registered: %s' % command)

        session.add(TesterModel(codec=self, command=command))
        session.commit()

    def unregister_tester(self, session, command):
        existing = session.query(TesterModel).filter(
            TesterModel.codec == self,
            TesterModel.command == command
        ).first()
        if not existing:
            raise SoundforestError('File tester was not registered: %s' % command)

        session.delete(existing)
        session.commit()


class ExtensionModel(Base):

    """ExtensionModel

    Filename extensions associated with audio format codecs

    """

    __tablename__ = 'extensions'

    id = Column(Integer, primary_key=True)
    extension = Column(SafeUnicode)
    codec_id = Column(Integer, ForeignKey('codecs.id'), nullable=False)
    codec = relationship('CodecModel',
        single_parent=False,
        backref=backref('extensions',
            order_by=extension,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return self.extension


class TesterModel(Base):
    """TesterModel

    Command to test audio files with given codec

    """
    __tablename__ = 'formattester'

    id = Column(Integer, primary_key=True)
    priority = Column(Integer)
    command = Column(SafeUnicode)

    codec_id = Column(Integer, ForeignKey('codecs.id'), nullable=False)
    codec = relationship('CodecModel',
        single_parent=False,
        backref=backref('testers',
            order_by=command,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return '%s format tester: %s' % (
            self.codec.name,
            self.command
        )


class DecoderModel(Base):

    """DecoderModel

    Audio format codec decoder commands

    """

    __tablename__ = 'decoders'

    id = Column(Integer, primary_key=True)
    priority = Column(Integer)
    command = Column(SafeUnicode)
    codec_id = Column(Integer, ForeignKey('codecs.id'), nullable=False)
    codec = relationship('CodecModel',
        single_parent=False,
        backref=backref('decoders',
            order_by=priority,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return '%s decoder: %s' % (
            self.codec.name,
            self.command
        )


class EncoderModel(Base):

    """EncoderModel

    Audio format codec encoder commands

    """

    __tablename__ = 'encoders'

    id = Column(Integer, primary_key=True)
    priority = Column(Integer)
    command = Column(SafeUnicode)
    codec_id = Column(Integer, ForeignKey('codecs.id'), nullable=False)
    codec = relationship('CodecModel',
        single_parent=False,
        backref=backref('encoders',
            order_by=priority,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return '%s encoder: %s' % (
            self.codec.name,
            self.command
        )


class PlaylistTreeModel(Base, BaseNamedModel):

    """PlaylistTreeModel

    PlaylistTreeModel parent folders

    """

    __tablename__ = 'playlist_trees'

    id = Column(Integer, primary_key=True)
    name = Column(SafeUnicode)
    path = Column(SafeUnicode)

    def __repr__(self):
        return '%s: %s' % (self.name, self.path)

    @property
    def exists(self):
        """Check if path exists

        Return true if registered path exists

        """
        return os.path.isdir(os.path.realpath(self.path))

    def update(self, session, source):
        """Read playlists to database from source

        Source must be iterable playlist object, for example
        soundforest.playlist.m3uPlaylistDirectory

        """
        for playlist in source:

            directory = os.path.realpath(playlist.directory)
            db_playlist = session.query(PlaylistModel).filter(
                PlaylistModel.parent == self,
                PlaylistModel.folder == directory,
                PlaylistModel.name == playlist.name,
                PlaylistModel.extension == playlist.extension
            ).first()

            if db_playlist is None:
                db_playlist = PlaylistModel(
                    parent=self,
                    folder=directory,
                    name=playlist.name,
                    extension=playlist.extension
                )
                session.add(db_playlist)

            for existing_track in db_playlist.tracks:
                session.delete(existing_track)

            try:
                playlist.read()
            except PlaylistError, emsg:
                logger.debug('Error reading playlist %s: %s' % (playlist, emsg))
                continue

            tracks = []
            for index, path in enumerate(playlist):
                position = index+1
                tracks.append(PlaylistTrackModel(
                    playlist=db_playlist,
                    path=path,
                    position=position
                ))
            session.add_all(tracks)
            db_playlist.updated = datetime.now()
            session.commit()


class PlaylistModel(Base, BaseNamedModel):

    """PlaylistModel

    PlaylistModel file of audio tracks

    """

    __tablename__ = 'playlists'

    id = Column(Integer, primary_key=True)

    updated = Column(Date)
    folder = Column(SafeUnicode)
    name = Column(SafeUnicode)
    extension = Column(SafeUnicode)
    description = Column(SafeUnicode)

    parent_id = Column(Integer, ForeignKey('playlist_trees.id'), nullable=False)
    parent = relationship('PlaylistTreeModel',
        single_parent=False,
        backref=backref('playlists',
            order_by=[folder, name],
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return '%s: %d tracks' % (
            os.sep.join([self.folder, self.name]),
            len(self.tracks)
        )

    def __len__(self):
        return len(self.tracks)


class PlaylistTrackModel(Base, BasePathNamedModel):

    """PlaylistTrackModel

    Audio track in a playlist

    """

    __tablename__ = 'playlist_tracks'

    id = Column(Integer, primary_key=True)

    position = Column(Integer)
    path = Column(SafeUnicode)

    playlist_id = Column(Integer, ForeignKey('playlists.id'), nullable=False)
    playlist = relationship('PlaylistModel',
        single_parent=False,
        backref=backref('tracks',
            order_by=position,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return '%d %s' % (self.position, self.path)


class TreeTypeModel(Base, BaseNamedModel):

    """TreeTypeModel

    Audio file tree types (music, samples, loops etc.)

    """

    __tablename__ = 'treetypes'

    id = Column(Integer, primary_key=True)
    name = Column(SafeUnicode)
    description = Column(SafeUnicode)

    def __repr__(self):
        return self.name


class PrefixModel(Base, BasePathNamedModel):

    """PrefixModel

    Audio tree prefix paths

    """

    __tablename__ = 'prefixes'

    id = Column(Integer, primary_key=True)
    path = Column(SafeUnicode, unique=True)

    def __repr__(self):
        return self.path


class TreeModel(Base, BasePathNamedModel):

    """TreeModel

    Audio file tree

    """

    __tablename__ = 'trees'

    id = Column(Integer, primary_key=True)
    path = Column(SafeUnicode, unique=True)
    source = Column(SafeUnicode, default=u'filesystem')
    description = Column(SafeUnicode)

    type_id = Column(Integer, ForeignKey('treetypes.id'), nullable=True)
    type = relationship('TreeTypeModel',
        single_parent=True,
        backref=backref('trees',
            order_by=path,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return self.path

    def album_count(self, session):
        return session.query(AlbumModel).filter(
            AlbumModel.tree == self
        ).count()

    def song_count(self, session):
        return session.query(TrackModel).filter(
            TrackModel.tree == self
        ).count()

    def tag_count(self, session):
        return session.query(TagModel)\
            .filter(TrackModel.tree == self)\
            .filter(TagModel.track_id == TrackModel.id)\
            .count()

    def match_tag(self, session, match):
        """Match database track tags

        Return tracks matching given tag value

        """
        return session.query(TrackModel)\
            .filter(TrackModel.tree == self)\
            .filter(TagModel.track_id == TrackModel.id)\
            .filter(TagModel.value.like('%%%s%%' % match))\
            .all()

    def filter_tracks(self, session, path):
        res = session.query(TrackModel).filter(TrackModel.tree == self)
        return res.filter(
            TrackModel.directory.like('%%%s%%' % path) |
            TrackModel.filename.like('%%%s%%' % path)
        ).all()

    def to_json(self):
        """Return tree as JSON

        Return tree path, description albums and total counters as JSON

        """
        album_info = [{'id': a.id, 'path': a.directory} for a in self.albums]

        return json.dumps({
            'id': self.id,
            'path': self.path,
            'description': self.description,
            'albums': album_info,
            'total_albums': len(self.albums),
            'total_songs': len(self.songs),
        })


class AlbumModel(Base, BasePathNamedModel):

    """AlbumModel

    AlbumModel of music tracks in tree database.

    """

    __tablename__ = 'albums'

    id = Column(Integer, primary_key=True)

    directory = Column(SafeUnicode)
    mtime = Column(Integer)
    tree_id = Column(Integer, ForeignKey('trees.id'), nullable=True)
    tree = relationship('TreeModel',
        single_parent=False,
        backref=backref('albums',
            order_by=directory,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return self.directory

    @property
    def path(self):
        return self.directory

    @property
    def relative_path(self):
        path = self.directory
        if self.tree and path[:len(self.tree.path)] == self.tree.path:
            path = path[len(self.tree.path):].lstrip(os.sep)
        return path

    @property
    def exists(self):
        return os.path.isdir(self.directory)

    @property
    def modified_isoformat(self, tz=None):
        if self.mtime is None:
            return None

        tval = datetime.fromtimestamp(self.mtime).replace(tzinfo=pytz.utc)

        if tz is not None:
            if isinstance(tz, basestring):
                tval = tval.astimezone(pytz.timezone(tz))
            else:
                tval = tval.astimezone(tz)

        return tval.isoformat()

    def to_json(self):
        """Return album as JSON

        Return album metadata + tracks IDs and filenames as JSON

        """
        track_info = [{'id': t.id, 'filename': t.filename} for t in self.tracks]
        return json.dumps({
            'id': self.id,
            'path': self.directory,
            'modified': self.modified_isoformat,
            'tracks': track_info
        })


class AlbumArtModel(Base):

    """AlbumArtModel

    AlbumArtModel files for music albums in tree database.

    """

    __tablename__ = 'albumarts'

    id = Column(Integer, primary_key=True)
    mtime = Column(Integer)
    albumart = Column(Base64Field)

    album_id = Column(Integer, ForeignKey('albums.id'), nullable=True)
    album = relationship('AlbumModel',
        single_parent=False,
        backref=backref('albumarts',
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return 'AlbumArtModel for %s' % self.album.path


class TrackModel(Base, BasePathNamedModel):

    """TrackModel

    Audio file. Optionally associated with a audio file tree

    """

    __tablename__ = 'tracks'

    id = Column(Integer, primary_key=True)

    directory = Column(SafeUnicode)
    filename = Column(SafeUnicode)
    extension = Column(SafeUnicode)
    checksum = Column(SafeUnicode)
    mtime = Column(Integer)
    deleted = Column(Boolean)

    tree_id = Column(Integer, ForeignKey('trees.id'), nullable=True)
    tree = relationship('TreeModel',
        single_parent=False,
        backref=backref('tracks',
            order_by=[directory, filename],
            cascade='all, delete, delete-orphan'
        )
    )
    album_id = Column(Integer, ForeignKey('albums.id'), nullable=True)
    album = relationship('AlbumModel',
        single_parent=False,
        backref=backref('tracks',
            order_by=[directory, filename],
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return os.sep.join([self.directory, self.filename])

    @property
    def path(self):
        return os.path.join(self.directory, self.filename)

    @property
    def relative_path(self):
        path = os.path.join(self.directory, self.filename)
        if self.tree and path[:len(self.tree.path)] == self.tree.path:
            path = path[len(self.tree.path):].lstrip(os.sep)
        return path

    @property
    def exists(self):
        return os.path.isfile(self.path)

    @property
    def modified_isoformat(self, tz=None):
        if self.mtime is None:
            return None

        tval = datetime.fromtimestamp(self.mtime).replace(tzinfo=pytz.utc)

        if tz is not None:
            if isinstance(tz, basestring):
                tval = tval.astimezone(pytz.timezone(tz))
            else:
                tval = tval.astimezone(tz)

        return tval.isoformat()

    def update(self, session, track):
        for tag in session.query(TagModel).filter(TagModel.track==self):
            session.delete(tag)
        for tag, value in track.tags.items():
            session.add(TagModel(track=self, tag=tag, value=value))
        session.commit()

    def to_json(self):
        return json.dumps({
            'id': self.id,
            'filename': self.path,
            'md5': self.checksum,
            'modified': self.modified_isoformat,
            'tags': dict((t.tag, t.value) for t in self.tags)
        })


class TagModel(Base):
    """TagModel

    Metadata tag for an audio file

    """

    __tablename__='tags'

    id=Column(Integer, primary_key = True)
    tag=Column(SafeUnicode)
    value=Column(SafeUnicode)
    base64_encoded=Column(Boolean)

    track_id=Column(Integer, ForeignKey('tracks.id'), nullable = False)
    track=relationship('TrackModel',
        single_parent = False,
        backref = backref('tags',
            order_by=tag,
            cascade='all, delete, delete-orphan'
        )
    )

    def __repr__(self):
        return '%s=%s' % (self.tag, self.value)


class SoundforestDB(object):

    """SoundforestDB

    Music database storing settings, synchronization data and music tree file metadata

    """

    def __init__(self, path=None, engine=None, debug=False):
        """
        By default, use sqlite databases in file given by path.
        """

        if engine is None:
            if path is None:
                path = DEFAULT_DATABASE

            config_dir = os.path.dirname(path)
            if not os.path.isdir(config_dir):
                try:
                    os.makedirs(config_dir)
                except OSError, (ecode, emsg):
                    raise SoundforestError('Error creating directory: %s' % config_dir)

            engine = create_engine('sqlite:///%s' % path, encoding='utf-8', echo=debug)

        event.listen(engine, 'connect', self._fk_pragma_on_connect)
        Base.metadata.create_all(engine)

        session_instance = sessionmaker(bind=engine)
        self.session = session_instance()

    def _fk_pragma_on_connect(self, connection, record):
        """Enable foreign keys for sqlite databases"""
        if isinstance(connection, SQLite3Connection):
            cursor = connection.cursor()
            cursor.execute('pragma foreign_keys=ON')
            cursor.close()

    def query(self, *args, **kwargs):
        """Wrapper to do a session query"""
        return self.session.query(*args, **kwargs)

    def rollback(self):
        """Wrapper to rolllback current session query"""
        return self.session.rollback()

    def commit(self):
        """Wrapper to commit current session query"""
        return self.session.commit()

    def as_dict(self, result):
        """Returns current query Base result as dictionary"""
        if not hasattr(result, '__table__'):
            raise ValueError('Not a sqlalchemy ORM result')
        return dict((k.name, getattr(result, k.name)) for k in result.__table__.columns)

    def add(self, items):
        """Add items in query session, committing changes"""
        if isinstance(items, list):
            self.session.add_all(items)
        else:
            self.session.add(items)

        self.session.commit()

    def delete(self, items):
        """Delete items in query session, committing changes"""
        if isinstance(items, list):
            for item in items:
                self.session.delete(item)
        else:
            self.session.delete(items)

        self.session.commit()

    @property
    def registered_sync_targets(self):
        return self.query(SyncTargetModel).all()

    @property
    def registered_codecs(self):
        return self.query(CodecModel).order_by(CodecModel.name).all()

    @property
    def registered_tree_types(self):
        return self.query(TreeTypeModel).order_by(TreeTypeModel.name).all()

    @property
    def registered_playlist_trees(self):
        """Return registered PlaylistTreeModel objects from database"""
        return self.query(PlaylistTreeModel).order_by(PlaylistTreeModel.name).all()

    @property
    def playlists(self):
        """Return registered PlaylistModel objects from database"""
        return self.query(PlaylistModel).all()

    @property
    def prefixes(self):
        """Return registered PrefixModel objects from database"""
        return self.query(PrefixModel).all()

    @property
    def trees(self):
        """Return registered TreeModel objects from database"""
        return self.query(TreeModel).all()

    @property
    def albums(self):
        """Return registered AlbumModel objects from database"""
        return self.query(AlbumModel).all()

    @property
    def tracks(self):
        """Return registered TrackModel objects from database"""
        return self.query(TrackModel).all()

    @property
    def registered_settings(self):
        """Return registered SettingModel objects from database"""
        return self.query(SettingModel).order_by(SettingModel.key).all()

    def update_setting(self, key, value):
        setting = self.query(SettingModel).filter(
            SettingModel.key == key
        ).first()

        if setting:
            existing.value = value
            self.update(existing)
        else:
            self.add(SettingModel(key=key, value=value))

    def get_setting(self, key):
        setting = self.query(SettingModel).filter(
            SettingModel.key == key
        ).first()
        if setting:
            return setting.value
        else:
            return None

    def register_sync_target(self, name, synctype, src, dst, flags=None, defaults=False):
        """Register a sync target"""

        existing = self.query(SyncTargetModel).filter(
            SyncTargetModel.name == name
        ).first()
        if existing:
            raise SoundforestError('Sync target was already registerd: %s' % name)

        target = SyncTargetModel(
            name=name,
            type=synctype,
            src=src,
            dst=dst,
            flags=flags,
            defaults=defaults
        )

        self.add(target)
        return target.as_dict()

    def unregister_sync_target(self, name):
        existing = self.query(SyncTargetModel).filter(SyncTargetModel.name == name).first()
        if not existing:
            raise SoundforestError('Sync target was not registered: %s' % name)

        self.delete(existing)

    def register_codec(self, name, extensions, description='', decoders=[], encoders=[], testers=[]):
        """
        Register codec with given parameters
        """
        codec = CodecModel(name=name, description=description)

        extension_instances = []
        for ext in extensions:
            extension_instances.append(ExtensionModel(codec=codec, extension=ext))

        decoder_instances = []
        for priority, command in enumerate(decoders):
            decoder_instances.append(DecoderModel(codec=codec, priority=priority, command=command))

        encoder_instances = []
        for priority, command in enumerate(encoders):
            encoder_instances.append(EncoderModel(codec=codec, priority=priority, command=command))

        tester_instances = []
        for priority, command in enumerate(testers):
            tester_instances.append(TesterModel(codec=codec, priority=priority, command=command))

        self.add([codec] + extension_instances + decoder_instances + encoder_instances + tester_instances)
        return codec

    def register_tree_type(self, name, description=''):
        existing = self.query(TreeTypeModel).filter(
            TreeTypeModel.name == name
        ).first()
        if existing:
            raise SoundforestError('Tree type was already registered: %s' % name)

        self.add(TreeTypeModel(name=name, description=description))

    def unregister_tree_type(self, name, description=''):
        existing = self.query(TreeTypeModel).filter(
            TreeTypeModel.name == name
        ).first()
        if not existing:
            raise SoundforestError('Tree type was not registered: %s' % name)

        self.delete(existing)

    def register_playlist_tree(self, path, name='Playlists'):
        existing = self.query(PlaylistTreeModel).filter(
            PlaylistTreeModel.path == path
        ).first()
        if existing:
            raise SoundforestError('Playlist source is already registered: %s' % path)

        self.add(PlaylistTreeModel(path=path, name=name))

    def unregister_playlist_tree(self, path):
        existing = self.query(PlaylistTreeModel).filter(
            PlaylistTreeModel.path == path
        ).first()
        if not existing:
            raise SoundforestError('Playlist source is not registered: %s' % path)

        self.delete(existing)

    def register_prefix(self, path):
        """Register a prefix"""
        if isinstance(path, str):
            path = unicode(path, 'utf-8')

        existing = self.query(PrefixModel).filter(
            PrefixModel.path == path
        ).first()
        if existing:
            raise SoundforestError('Prefix was already registered: %s' % path)

        self.add(PrefixModel(path=path))

    def unregister_prefix(self, path):
        """Unregister a prefix"""
        existing = self.query(PrefixModel).filter(
            PrefixModel.path == path
        ).first()
        if not existing:
            raise SoundforestError('Prefix was not registered: %s' % path)

        self.delete(existing)

    def register_tree(self, path, description='', tree_type='songs'):
        """Register tree"""
        if isinstance(path, str):
            path = unicode(path, 'utf-8')

        existing = self.query(TreeModel).filter(
            TreeModel.path == path
        ).first()
        if existing:
            raise SoundforestError('Tree was already registered: %s' % path)

        tt = self.get_tree_type(tree_type)
        self.add(TreeModel(path=path, description=description, type=tt))

    def unregister_tree(self, path, description=''):
        """Unregister tree"""
        existing = self.query(TreeModel).filter(
            TreeModel.path == path
        ).first()
        if not existing:
            raise SoundforestError('Tree was not registered: %s' % path)

        self.delete(existing)

    def get_codec(self, name):
        """Return codec matching name"""
        return self.query(CodecModel).filter(
            CodecModel.name == name
        ).first()

    def get_tree_type(self, name):
        """Return tree type matching name"""
        return self.query(TreeTypeModel).filter(
            TreeTypeModel.name == name
        ).first()

    def get_tree(self, path, tree_type='songs'):
        """Return tree matching path"""
        return self.query(TreeModel).filter(
            TreeModel.path == path
        ).first()

    def get_album(self, path):
        """Return album matching path"""
        return self.query(AlbumModel).filter(
            AlbumModel.directory == path
        ).first()

    def get_track(self, path):
        """Return trach matching path"""
        return self.query(TrackModel).filter(
            TrackModel.directory == os.path.dirname(path),
            TrackModel.filename == os.path.basename(path),
        ).first()

    def get_playlist_tree(self, path):
        return self.query(PlaylistTreeModel).filter(
            PlaylistTreeModel.path == path
        ).first()

    def get_playlist(self, path):
        """Get playlist by path"""
        return self.query(PlaylistModel).filter(
            PlaylistModel.path == path
        ).first()



# coding=utf-8
"""Tag abstraction

Tag metadata reader and writer classes

"""

import os
import base64
import json
import logging
from datetime import datetime

from soundforest import normalized
from soundforest.log import SoundforestLogger
from soundforest.formats import AudioFileFormat
from soundforest.tags import TagError
from soundforest.tags.constants import STANDARD_TAG_ORDER, STANDARD_TAG_MAP
from soundforest.tags.xmltag import XMLTags, XMLTagError
from soundforest.tags.albumart import AlbumArt, AlbumArtError

__all__ = (
    'aac',
    'flac',
    'mp3',
    'vorbis'
)

YEAR_FORMATTERS = [
    lambda x: unicode('%s'% int(x), 'utf-8'),
    lambda x: unicode('%s'% datetime.strptime(x, '%Y-%m-%d').strftime('%Y'), 'utf-8'),
    lambda x: unicode('%s'% datetime.strptime(x, '%Y-%m-%dT%H:%M:%SZ').strftime('%Y'), 'utf-8'),
]

logger = SoundforestLogger().default_stream

class TagParser(dict):
    """
    Parent class for tag parser implementations
    """

    def __init__(self, codec, path, tag_map=None):
        dict.__init__(self)
        self.codec = codec
        self.path = normalized(os.path.realpath(path))
        self.tag_map = tag_map is not None and tag_map or {}
        self.entry = None
        self.modified = False

        self.albumart_obj = None
        self.supports_albumart = False

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError('No such TagParser attribute: %s' % attr)

    def __getitem__(self, item):
        """
        Return tags formatted to unicode, decimal.Decimal or
        other supported types.
        Does not include albumart images, which are accessed
        via self.albumart attribute
        """
        fields = self.__tag2fields__(item)
        for field in fields:
            if field not in self.entry:
                continue

            tag = self.entry[field]
            if not isinstance(tag, list):
                tag = [tag]

            values = []
            for value in tag:
                if not isinstance(value, unicode):
                    if isinstance(value, int):
                        value = unicode('%d' % value)

                    else:
                        try:
                            value = unicode(value, 'utf-8')
                        except UnicodeDecodeError, emsg:
                            raise TagError('Error decoding %s tag %s: %s' % (self.path, field, emsg))

                values.append(value)

            return values

        raise KeyError('No such tag: %s' % fields)

    def __setitem__(self, item, value):
        if isinstance(item, AlbumArt):
            try:
                self.albumart_obj.import_albumart(value)
            except AlbumArtError, emsg:
                raise TagError('Error setting albumart: %s' % emsg)

        self.set_tag(item, value)

    def __delitem__(self, item):
        fields = self.__tag2fields__(item)
        for tag in fields:
            if tag not in self.keys():
                continue

            del self.entry[tag]
            self.modified = True

    def __tag2fields__(self, tag):
        """
        Resolve tag name to internal parser field
        """
        for name, tags in self.tag_map.items():
            if tag == name:
                return tags

        return [tag]

    def __field2tag__(self, field):
        """
        Resolve internal parser field to tag name
        """
        for name, tags in self.tag_map.items():
            # Can happen if name is internal reference: ignore here
            if tags is None:
                continue

            if field in tags:
                return name

        return field

    def __normalized_tag__(self, tag):
        """
        Return list of values for tag, mangling to standard formats

        Returns None for non-string (coverart etc.) tags
        """
        formatted = []
        try:
            tags = self[tag]
        except KeyError:
            return None

        if not isinstance(tags, list):
            tags = [tags]

        if len(tags) == 0:
            return None

        for value in tags:
            if not isinstance(value, basestring):
                continue

            if tag=='year':
                # Try to clear extra date details from year
                for fmt in YEAR_FORMATTERS:
                    try:
                        value = fmt(value)
                        break
                    except ValueError, emsg:
                        pass

            formatted.append(value)

        if not formatted:
            return None

        return formatted

    def __tagname__(self, tag):
        if tag in STANDARD_TAG_MAP:
            return STANDARD_TAG_MAP[tag]['label']
        return tag

    def __repr__(self):
        return '%s: %s' % (self.codec, self.path)

    @property
    def mtime(self):
        return os.stat(self.path).st_mtime

    @property
    def atime(self):
        return os.stat(self.path).st_atime

    @property
    def ctime(self):
        return os.stat(self.path).st_ctime

    @property
    def albumart(self):
        if not self.supports_albumart or not self.albumart_obj:
            return None
        return self.albumart_obj

    def set_albumart(self, albumart):
        if not self.supports_albumart:
            raise TagError('Format does not support albumart')
        return self.albumart.import_albumart(albumart)

    def remove_tag(self, item):
        if not self.has_key(item):
            raise TagError('No such tag: %s' % item)
        del self[tag]

    def get_tag(self, item):
        """
        Return tag from file. Raises TagError if tag is not found.
        """
        if not self.has_key(item):
            raise TagError('No such tag: %s' % item)

        value = self.__normalized_tag__(item)
        if value is None:
            raise TagError('No such string tag: %s' % item)

        return value

    def set_tag(self, item, value):
        """
        Sets the tag item to given value.
        Must be implemented in child class, this raises
        NotImplementedError
        """
        raise NotImplementedError('Must implement set_tag in child')

    def get_raw_tags(self):
        """
        Get internal presentation of tags
        """
        return self.entry.items()

    def get_unknown_tags(self):
        """
        Must be implemented in child if needed: return empty list here
        """
        return []

    def sort_keys(self, keys):
        """
        Sort keys with STANDARD_TAG_ORDER list
        """
        values = []
        for k in STANDARD_TAG_ORDER:
            if k in keys:
                values.append(k)

        for k in keys:
            if k not in STANDARD_TAG_ORDER:
                values.append(k)

        return values

    def has_key(self, key):
        """
        Test if given key is in tags
        """
        keys = self.__tag2fields__(key)
        for k in keys:
            if k in self.keys():
                return True

        return False

    def keys(self):
        """
        Return file tag keys mapped with tag_map.
        """
        return self.sort_keys(
            [self.__field2tag__(k) for k in self.entry.keys()]
        )

    def items(self):
        """
        Return tag, value pairs using tag_map keys.
        """
        tags = []
        for tag in self.keys():
            values = self.__normalized_tag__(tag)

            if values is None:
                continue

            tags.append((tag, values))

        return tags

    def values(self):
        """
        Return tag values from entry.
        """
        tags = []
        for tag in self.keys():
            values = self.__normalized_tag__(tag)

            if values is None:
                continue

            tags.append(value)

        return tags

        return [self[k] for k, v in self.keys()]

    def as_dict(self):
        """
        Return tags as dictionary
        """
        return dict(self.items())

    def as_xml(self):
        """
        Return tags formatted as XML
        """
        return XMLTags(self.as_dict())

    def to_json(self, indent=2):
        """
        Return tags formatted as json
        """
        stat = os.stat(self.path)
        return json.dumps(
            {
                'filename': self.path,
                'modified': datetime.fromtimestamp(stat.st_mtime).isoformat(),
                'size': stat.st_size,
                'tags': [{'tag': k, 'name': self.__tagname__(k), 'values': v} for k, v in self.items()]
            },
            ensure_ascii=False,
            indent=indent,
            sort_keys=True
        )

    def get_unknown_tags(self):
        """
        Must be implemented in child if needed: return empty list here
        """
        return []

    def update_tags(self, data):
        if not isinstance(data, dict):
            raise TagError('Updated tags must be a dictionary instance')

        for k, v in data.items():
            self.set_tag(k, v)

        return self.modified

    def replace_tags(self, data):
        """Replace tags

        Set tags given in data dictionary, removing any existing tags.
        """
        if not isinstance(data, dict):
            raise TagError('Updated tags must be a dictionary instance')

        self.clear_tags()

        return self.update_tags(data)

    def remove_tags(self, tags):
        """
        Remove given list of tags from file
        """
        for tag in tags:
            del self[tag]

        if self.modified:
            self.save()

    def clear_tags(self):
        """
        Remove all tags from file
        """
        self.entry.clear()
        self.save()

    def remove_unknown_tags(self):
        """
        Remove any tags which we don't know about.
        """
        for tag in self.unknown_tags:
            del self[tag]

        if self.modified:
            self.save()

    def save(self):
        """
        Save tags to file.
        """
        try:
            for attr in ('track_numbering', 'disk_numbering'):
                try:
                    tag = getattr(self, attr)
                    tag.save_tag()
                except ValueError, emsg:
                    logger.debug('Error processing %s: %s' % (attr, emsg))

            if not self.modified:
                return

            self.entry.save()

        except OSError, (ecode, emsg):
            raise TagError(emsg)

        except IOError, (ecode, emsg):
            raise TagError(emsg)

class TrackAlbumart(object):
    """
    Parent class for common albumart operations
    """

    def __init__(self, track):
        self.track = track
        self.modified = False
        self.albumart = None

    def __repr__(self):
        return self.albumart.__repr__()

    @property
    def info(self):
        if self.albumart is None:
            return {}
        self.albumart.get_info()

    @property
    def defined(self):
        """
        Returns True if albumart is defined, False otherwise
        """
        if self.albumart is None:
            return False
        return True

    def as_base64_tag(self):
        """
        Return albumart image data as base64_tag tag
        """
        if self.albumart is None:
            raise TagError('Albumart is not loaded')
        return base64_tag(base64.b64encode(self.albumart.dump()))

    def import_albumart(self, albumart):
        """
        Parent method to set albumart tag. Child class must
        implement actual embedding of the tag to file.
        """
        if not isinstance(albumart, AlbumArt):
            raise TagError('Albumart must be AlbumArt instance')
        if not albumart.is_loaded:
            raise TagError('Albumart to import is not loaded with image.')
        self.albumart = albumart

    def save(self, path):
        """
        Save current albumart to given pathname
        """
        if self.albumart is None:
            raise TagError('Error saving albumart: albumart is not loaded')
        self.albumart.save(path)


class TrackNumberingTag(object):
    """
    Parent class for processing track numbering info, including track and
    disk numbers and total counts.

    Fields should be set and read from attributes 'value' and 'total'
    """

    def __init__(self, track, tag):
        self.track = track
        self.tag = tag
        self.f_value = None
        self.f_total = None

    def __repr__(self):
        if self.total is not None:
            return '%d/%d' % (self.value, self.total)

        elif self.value is not None:
            return '%d' % (self.value)

        else:
            return None

    def __getattr__(self, attr):
        if attr == 'value':
            return self.f_value

        if attr == 'total':
            return self.f_total

        raise AttributeError('No such TrackNumberingTag attribute: %s' % attr)

    def __setattr__(self, attr, value):
        if attr in ['value', 'total']:
            if isinstance(value, list):
                value = value[0]

            try:
                if value is not None:
                    value = int(value)
            except ValueError:
                raise TagError('TrackNumberingTag values must be integers')
            except TypeError:
                raise TagError('TrackNumberingTag values must be integers')

            if attr == 'value':
                self.f_value = value

            if attr == 'total':
                self.f_total = value

        else:
            object.__setattr__(self, attr, value)

    def save_tag(self):
        """
        Export this numbering information back to file tags.

        If value is None, ignore both values without setting tag.
        If total is None but value is set, set total==value.
        """
        raise NotImplementedError('save_tag must be implemented in child class')

def Tags(path, fileformat=None):
    """
    Loader for file metadata tags. Tag reading and writing for various
    file formats is implemented by tag formatter classes in module
    soundforest.tags.formats, initialized automatically by this class.
    """
    if not os.path.isfile(path):
        raise TagError('No such file: %s' % path)

    path = normalized(os.path.realpath(path))

    if fileformat is None:
        fileformat = AudioFileFormat(path)

    if not isinstance(fileformat, AudioFileFormat):
        raise TagError('File format must be AudioFileFormat instance')

    fileformat = fileformat
    if fileformat.is_metadata:
        raise TagError('Attempting to load audio tags from metadata file')

    if fileformat.codec is None:
        raise TagError('Unsupported audio file: %s' % path)

    tag_parser = fileformat.get_tag_parser()
    if tag_parser is None:
        return None

    return tag_parser(fileformat.codec, path)



#!/usr/bin/env python

import os
import codecs
import logging

from soundforest import normalized

class PlaylistError(Exception):
    pass


class Playlist(list):
    def __init__(self, name, unique=True):
        self.name = os.path.splitext(os.path.basename(name))[0]
        self.unique = unique
        self.modified = False
        self.path = None

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.path

    def __len__(self):
        if list.__len__(self) == 0:
            self.read()
        return list.__len__(self)

    @property
    def exists(self):
        return os.path.isfile(self.path)

    def read(self):
        raise NotImplementedError('You must implement reading in subclass')

    def write(self):
        raise NotImplementedError('You must implement writing in subclass')

    def __insert(self, path, position=None):
        if self.unique and self.count(path) > 0:
            return

        self.modified = True
        if not position:
            list.append(self, path)

        else:
            try:
                position = int(position)
                if position < 0:
                    raise ValueError

                if position > list.__len__(self):
                    position = list.__len__(self)

            except ValueError:
                raise PlaylistError('Invalid position: %s' % position)

            self.insert(position, path)

    def append(self, path, position=None, recursive=False):
        path = normalized(os.path.realpath(path))
        if os.path.isfile(path):
            self.__insert(path, position)

        elif os.path.isdir(path):
            for f in ['%s' % os.path.join(path, x) for x in os.listdir(path)]:
                f = normalized(os.path.realpath(f))

                if not recursive and os.path.isdir(f):
                    continue

                self.append(f, position)

        else:
            raise PlaylistError('Not a file or directory: %s' % path)

        self.modified = True

class m3uPlaylist(Playlist):
    def __init__(self, name, config=None, folder=None, unique=True):
        Playlist.__init__(self, name, unique)

        if os.path.isfile(name):
            path = os.path.realpath(name)

        else:
            if folder is not None:
                path = os.path.join(folder, '%s.m3u' % self.name)
            else:
                path = os.path.join('%s.m3u' % self.name)

        self.path = normalized(os.path.realpath(path))
        self.filename = os.path.basename(self.path)
        self.folder = os.path.dirname(self.path)

    def read(self):
        if not self.exists:
            return

        try:
            with open(self.path, 'r') as lines:
                self.__delslice__(0, list.__len__(self))
                for l in lines:
                    l = l.strip()
                    if l.startswith('#'):
                        continue

                    filepath = normalized(os.path.realpath(l))
                    if not os.path.isfile(filepath):
                        continue

                    if self.unique and self.count(filepath)>0:
                        continue

                    self.append(filepath)
        except IOError, (ecode, emsg):
            raise PlaylistError('Error reading %s: %s' % (self.path, emsg))

    def write(self):
        pl_dir = os.path.dirname(self.path)

        if not os.path.isdir(pl_dir):
            try:
                os.makedirs(pl_dir)

            except OSError, (ecode, emsg):
                raise PlaylistError('Error creating directory %s: %s' % pl_dir)

            except IOError, (ecode, emsg):
                raise PlaylistError('Error creating directory %s: %s' % pl_dir)

        if not self.modified:
            return

        try:
            fd = open(self.path, 'w')
            for f in self: fd.write('%s\n' % f)
            fd.close()

        except IOError, (ecode, emsg):
            raise PlaylistError('Error writing playlist %s: %s' % (self.path, emsg))

        except OSError, (ecode, emsg):
            raise PlaylistError('Error writing playlist %s: %s' % (self.path, emsg))

    def remove(self):
        if not os.path.isfile(self.path):
            return

        try:
            os.unlink(self.path)

        except OSError, (ecode, emsg):
            raise PlaylistError('Error removing playlist %s: %s' % (self.path, emsg))

        except oError, (ecode, emsg):
            raise PlaylistError('Error removing playlist %s: %s' % (self.path, emsg))


class m3uPlaylistDirectory(list):
    def __init__(self, path=None):
        self.path = path
        if not os.path.isdir(self.path):
            raise PlaylistError('No such directory: %s' % self.path)

        for f in sorted(os.listdir(self.path)):
            f = os.path.join(self.path, f)

            if os.path.isdir(f):
                self.extend(m3uPlaylistDirectory(path=f))
                continue

            if os.path.splitext(f)[1][1:] != 'm3u':
                continue

            self.append(m3uPlaylist(f))

    def __getitem__(self, item):
        try:
            return self[int(item)]
        except IndexError:
            pass
        except ValueError:
            pass

        item = str(item)
        if str(item[-4:]) != '.m3u':
            item += '.m3u'

        try:
            name_lower = pl.filename.lower()
            return [item for item in self if item.lower() == name_lower][0]
        except IndexError:
            pass

        try:
            path = os.path.realpath(item)
            return [pl for pl in self if pl.path == path][0]
        except IndexError:
            pass

        raise IndexError('Invalid m3uPlaylistDirectory index %s' % item)




# coding=utf-8
"""FLAC tags

Flac file tag parser

"""

from mutagen.flac import FLAC
from mutagen.flac import Picture, FLACNoHeaderError

from soundforest.tags import TagError
from soundforest.tags.albumart import AlbumArt, AlbumArtError
from soundforest.tags.constants import OGG_MULTIPLE_VALUES_TAGS
from soundforest.tags.tagparser import TagParser, TrackNumberingTag, TrackAlbumart

FLAC_ALBUMART_TAG = 'METADATA_BLOCK_PICTURE'

FLAC_STANDARD_TAGS = {
    'album_artist':         ['ALBUM_ARTIST'],
    'artist':               ['ARTIST'],
    'arranger':             ['ARRANGER'],
    'author':               ['AUTHOR'],
    'composer':             ['COMPOSER'],
    'conductor':            ['CONDUCTOR'],
    'ensemble':             ['ENSEMBLE'],
    'orchestra':            ['ORCHESTRA'],
    'performer':            ['PERFORMER'],
    'publisher':            ['PUBLISHER'],
    'lyricist':             ['LYRICIST'],
    'album':                ['ALBUM'],
    'title':                ['TITLE'],
    'partnumber':           ['PARTNUMBER'],
    'tracknumber':          ['TRACKNUMBER'],
    'disknumber':           ['DISKNUMBER'],
    'genre':                ['GENRE'],
    'comment':              ['COMMENT'],
    'note':                 ['NOTE'],
    'description':          ['DESCRIPTION'],
    'location':             ['LOCATION'],
    'year':                 ['DATE'],
    'bpm':                  ['BPM'],
    'rating':               ['RATING'],
    'label':                ['LABEL'],
    'labelno':              ['LABELNO'],
    'opus':                 ['OPUS'],
    'isrc':                 ['ISRC'],
    'ean':                  ['EAN/UPN'],
    'lyrics':               ['LYRICS'],
    'website':              ['WEBSITE'],
    'copyright':            ['COPYRIGHT'],
    'version':              ['VERSION'],
    'sourcemedia':          ['SOURCEMEDIA'],
    'encoding':             ['ENCODING'],
    'encoded_by':           ['ENCODED-BY'],
    'sort_album_artist':    ['SORT_ALBUM_ARTIST'],
    'sort_artist':          ['SORT_ARTIST'],
    'sort_composer':        ['SORT_COMPOSER'],
    'sort_performer':       ['SORT_PERFORMER'],
    'sort_show':            ['SORT_SHOW'],
    'sort_album':           ['SORT_ALBUM'],
    'sort_title':           ['SORT_TITLE'],
}

FLAC_REPLAYGAIN_TAGS = {
    'album_gain':           ['REPLAYGAIN_ALBUM_GAIN'],
    'album_peak':           ['REPLAYGAIN_ALBUM_PEAK'],
    'track_gain':           ['REPLAYGAIN_TRACK_GAIN'],
    'track_peak':           ['REPLAYGAIN_TRACK_PEAK'],
}

FLAC_TAG_FORMATTERS = {

}

FLAC_EXTRA_TAGS = {
}

class FLACAlbumart(TrackAlbumart):
    """
    Encoding of flac albumart to flac Picture tags
    """
    def __init__(self, track):
        if not isinstance(track, flac):
            raise TagError('Track is not instance of flac')
        TrackAlbumart.__init__(self, track)

        try:
            self.albumart = AlbumArt()
            self.albumart.import_data(self.track.entry.pictures[0].data)
        except IndexError:
            self.albumart = None
            return

    def import_albumart(self, albumart):
        """
        Imports albumart object to the file tags.

        Sets self.track.modified to True
        """
        TrackAlbumart.import_albumart(self, albumart)

        p = Picture()
        [setattr(p, k, v) for k, v in self.albumart.info.items()]
        self.track.entry.add_picture(p)
        self.track.modified = True

class FLACNumberingTag(TrackNumberingTag):
    """
    FLAC tags for storing track or disk numbers.
    The tag can be either a single number or two numbers separated by /
    If total is given, the value must be integer.
    """
    def __init__(self, track, tag):
        TrackNumberingTag.__init__(self, track, tag)

        if not self.track.entry.has_key(self.tag):
            return

        value = self.track.entry[self.tag]
        try:
            value, total = value[0].split('/', 1)
        except ValueError:
            value = value[0]
            total = None
        self.value = value
        self.total = total

    def save_tag(self):
        """
        Set new numbering information to vorbis tags, marking file
        dirty to require saving but not saving tags.
        """
        value = self.__repr__()
        if value is not None:
            self.track.entry[self.tag] = '%s' % value
            self.track.modified = True

class flac(TagParser):
    """
    Class for processing Ogg FLAC file tags
    """
    def __init__(self, codec, path):
        TagParser.__init__(self, codec, path, tag_map=FLAC_STANDARD_TAGS)

        try:
            self.entry = FLAC(path)
        except IOError, emsg:
            raise TagError('Error opening %s: %s' % (path, str(emsg)))
        except FLACNoHeaderError, emsg:
            raise TagError('Error opening %s: %s' % (path, str(emsg)))

        self.albumart_obj = None
        self.track_numbering = FLACNumberingTag(self, 'TRACKNUMBER')
        self.disk_numbering = FLACNumberingTag(self, 'DISKNUMBER')

    def __getitem__(self, item):
        if item == 'tracknumber':
            return [unicode('%d' % self.track_numbering.value)]
        if item == 'totaltracks':
            return [unicode('%d' % self.track_numbering.total)]
        if item == 'disknumber':
            return [unicode('%d' % self.disk_numbering.value)]
        if item == 'totaldisks':
            return [unicode('%d' % self.disk_numbering.total)]
        return TagParser.__getitem__(self, item)

    def __delitem__(self, item):
        try:
            item, value = item.split('=', 1)
        except ValueError:
            value = None

        fields = self.__tag2fields__(item)
        for tag in fields:
            if tag not in self.entry.keys():
                continue

            del self.entry[tag]
            self.modified = True

    def __field2tag__(self, field):
        return TagParser.__field2tag__(self, field.upper())

    def keys(self):
        """
        Return tag names sorted with self.sort_keys()
        """
        keys = TagParser.keys(self)
        if 'TOTALTRACKS' in keys:
            keys.remove('TOTALTRACKS')
        if 'TOTALDISKS' in keys:
            keys.remove('TOTALDISKS')
        if 'TRACKNUMBER' in [x.upper() for x in keys]:
            if self.track_numbering.total is not None:
                keys.append('totaltracks')
        if 'DISKNUMBER' in [x.upper() for x in keys]:
            if self.disk_numbering.total is not None:
                keys.append('totaldisks')
        if FLAC_ALBUMART_TAG in [x.upper() for x in keys]:
            keys.remove(FLAC_ALBUMART_TAG)
        for replaygain_tag_fields in FLAC_REPLAYGAIN_TAGS.values():
            for tag in replaygain_tag_fields:
                if tag in keys:
                    keys.remove(tag)
        return [x.lower() for x in self.sort_keys(keys)]

    def has_key(self, tag):
        return tag.lower() in self.keys()

    def set_tag(self, item, value):
        """
        All flac tags are unicode strings, and there can be multiple
        tags with same name.
        We do special precessing for track and disk numbering.
        """
        if item == 'tracknumber':
            self.track_numbering.value = value
            return
        if item == 'totaltracks':
            self.track_numbering.total = value
            return
        if item == 'disknumber':
            self.disk_numbering.value = value
            return
        if item == 'totaldisks':
            self.disk_numbering.total = value
            return

        if not isinstance(value, list):
            value = [value]

        tags = self.__tag2fields__(item)
        item = tags[0]

        for tag in tags:
            if self.entry.has_key(tag):
                if tag in OGG_MULTIPLE_VALUES_TAGS and value not in self.entry[tag]:
                    value = self.entry[tag] + value

                del self.entry[tag]

        entries =[]
        for v in value:
            if FLAC_TAG_FORMATTERS.has_key(item):
                entries.append(FLAC_TAG_FORMATTERS[item](v))
            else:
                if not isinstance(v, unicode):
                    v = unicode(v, 'utf-8')
                entries.append(v)
        self.entry[item] = entries
        self.modified = True



#!/usr/bin/env python
"""
Setup for soundforest package for setuptools
"""

import os
import glob
from setuptools import setup, find_packages

VERSION ='3.4.2'

setup(
    name = 'soundforest',
    keywords = 'Sound Audio File Tree Codec Database',
    description = 'Audio file library manager',
    version = VERSION,
    author = 'Ilkka Tuohela',
    author_email = 'hile@iki.fi',
    license = 'PSF',
    url = 'http://tuohela.net/packages/soundforest',
    packages = find_packages(),
    scripts = glob.glob('bin/*'),
    install_requires = (
        'setproctitle',
        'sqlalchemy',
        'requests',
        'lxml',
        'pytz',
        'mutagen',
        'pillow',
    ),
)



"""
Common metadata file matches seen in music file directories.

Please note these classes are not used to open or process the metadata files:
this module is intended only for detecting the type of files.
"""

import os

# List of filenames we consider as albumart images. Can't be arsed to store this
# to database, it's a static list anyway and we can extend the list if needed!
ALBUMART_FILENAMES = (
    'artwork.jpg',
    'albumart.jpg',
    'coverart.jpg',
    'front.jpg',
    'back.jpg',
)

# Same for cover booklet filenames: rename these consistently in your library!
BOOKLET_FILENAMES = (
    'booklet.pdf',
)

# Apple OS/X system files we wish to ignore
OSX_SYSTEM_FILES = [
    '.com.apple.timemachine.supported',
    '.DS_Store',
]


class MetadataFileType(object):
    """
    Parent class for metadata file description classes.

    Attributes:
    description     Textual description for this metadata file type
    filenames       List of full filenames to match to this type of metadata
    extensions      List of file extensions to match to this type of metadata
    """
    def __init__(self, description, filenames=None, extensions=None):
        self.description = description
        self.filenames = filenames
        self.extensions = extensions

    def __repr__(self):
        """
        Returns the description string
        """
        return self.description

    def match(self, path):
        """
        Match given path to this metadata file type. Override in child class
        if you need more complicated logic.

        Returns true if the filename matches metadata type, False if not.
        """
        path = os.path.realpath(path)
        if self.filenames:
            if os.path.basename(path) in self.filenames:
                return True

        if self.extensions:
            ext = os.path.splitext(path)[1][1:].lower()
            if ext in self.extensions:
                return True

        return False


class OSXSystemFile(MetadataFileType):
    """
    OS/X system metadata files not relevant for audio trees.
    """
    def __init__(self):
        MetadataFileType.__init__(self, 'OS/X System file', filenames=OSX_SYSTEM_FILES)


class AbletonAnalysisFile(MetadataFileType):
    """
    Ableton track metadata files.
    """
    def __init__(self):
        MetadataFileType.__init__(self, 'Ableton Live Track Metadata', extensions=['asd'])


class PDFBooklet(MetadataFileType):
    """
    PDF format album booklet, as received from itunes.

    Right now, we expect the file to be renamed to booklet.pdf in same directory
    as the album. Someone else may add parser for PDF files in general if needed.
    """
    def __init__(self):
        MetadataFileType.__init__(self, 'Album Cover Booklet', filenames=BOOKLET_FILENAMES)


class CoverArt(MetadataFileType):
    """
    Coverart files stored to the album directory with music files.

    Static list of albumart filenames we process are defined in module sources.
    """
    def __init__(self):
        MetadataFileType.__init__(self, 'Album Artwork', filenames=ALBUMART_FILENAMES)


class m3uPlaylist(MetadataFileType):
    """
    Playlist files in m3u format
    """
    def __init__(self):
        MetadataFileType.__init__(self, 'm3u playlist', extensions=['m3u'])


class Metadata(list):
    """
    Load metadata parsers and match filenames to the parsers
    """
    def __init__(self):
        """
        Register instances of the MetadataFileType classes in the module
        """
        list.__init__(self)
        self.register_metadata(CoverArt())
        self.register_metadata(m3uPlaylist())
        self.register_metadata(AbletonAnalysisFile())
        self.register_metadata(PDFBooklet())
        self.register_metadata(OSXSystemFile())

    def register_metadata(self, metadata_class):
        """
        Register instance of a MetadataFileType class
        """
        if not isinstance(metadata_class, MetadataFileType):
            raise ValueError('Not a MetadataFileType instance')

        self.append(metadata_class)

    def match(self, path):
        """
        Match path to registered metadata type parsers.
        Returns matching metadata class or None
        """
        for m in self:
            if m.match(path):
                return m

        return None




# coding=utf-8
"""Tree prefixes

Tree prefixes configuration

"""

import os
import configobj

from soundforest.defaults import SOUNDFOREST_USER_DIR
from soundforest.config import ConfigDB
from soundforest.log import SoundforestLogger
from soundforest.formats import match_codec, path_string

USER_PATH_CONFIG = os.path.join(SOUNDFOREST_USER_DIR, 'paths.conf')

DEFAULT_PATHS = [
    '/music',
    '/Volumes/Media',
    os.path.join(os.getenv('HOME'), 'Music')
]

ITUNES_MUSIC = os.path.join(os.getenv('HOME'), 'Music', 'iTunes', 'iTunes Media', 'Music')
ITUNES_PARTS = ITUNES_MUSIC.split(os.sep)
for i in range(0, len(ITUNES_PARTS) + 1):
    if os.path.islink(os.sep.join(ITUNES_PARTS[:i])):
        ITUNES_MUSIC = os.path.realpath(ITUNES_MUSIC)
        break


class PrefixError(Exception): pass


class MusicTreePrefix(object):

    """MusicTreePrefix

    Tree path prefix matcher

    """
    def __init__(self, path, extensions=[]):
        self.log = SoundforestLogger().default_stream
        self.path = path.rstrip(os.sep)

        if not isinstance(extensions, list):
            raise PrefixError('Extensions must be a list')

        self.extensions = []
        for ext in extensions:
            if hasattr(ext, 'extension'):
                ext = ext.extension

            elif not isinstance(ext, basestring):
                raise PrefixError('Extensions must be a list of strings')

            if not isinstance(ext, unicode):
                ext = unicode(ext, 'utf-8')

            if self.extensions.count(ext)==0:
                self.extensions.append(ext)

    def __repr__(self):
        if self.extensions:
            return '%s (%s)' % (self.path, ','.join(self.extensions))
        else:
            return self.path

    @property
    def realpath(self):
        return os.path.realpath(self.path)

    def match(self, path):
        if path[:len(self.path)] == self.path:
            return True

        realpath = os.path.realpath(path)
        mypath = os.path.realpath(self.path)
        if realpath[:len(mypath)] == mypath:
            return True

        return False

    def match_extension(self, extension):
        if not isinstance(extension, unicode):
            extension = unicode(extension, 'utf-8')
        return extension in self.extensions

    def relative_path(self, path):
        path = path.rstrip(os.sep)
        if path[:len(self.path)] == self.path:
            return path_string(path[len(self.path):].lstrip(os.sep))

        realpath = os.path.realpath(path)
        mypath = os.path.realpath(self.path)

        if realpath[:len(mypath)] == mypath:
            return path_string(realpath[len(mypath):].lstrip(os.sep))

        raise PrefixError('Prefix does not match: %s' % path)


class TreePrefixes(object):

    """TreePrefixes

    List of known or common music tree prefixes

    """
    __instance = None

    def __init__(self):
        if TreePrefixes.__instance is None:
            TreePrefixes.__instance = TreePrefixes.TreePrefixInstance()
            self.__dict__['TreePrefixes.__instance'] = TreePrefixes.__instance

    class TreePrefixInstance(list):

        def __init__(self):
            self.log = SoundforestLogger().default_stream
            list.__init__(self)
            self.db = ConfigDB()

            common_prefixes = set(DEFAULT_PATHS + [prefix.path for prefix in self.db.prefixes])

            for path in common_prefixes:
                for name, codec in self.db.codecs.items():
                    prefix_path = os.path.join(path, name)
                    prefix = MusicTreePrefix(prefix_path, [codec.name] + codec.extensions)
                    self.register_prefix(prefix)

                if 'm4a' in self.db.codecs.keys():
                    prefix_path = os.path.join(path, 'm4a')
                    prefix = MusicTreePrefix(prefix_path, self.db.codecs.extensions('m4a'))
                    self.register_prefix(prefix)

            itunes_prefix = MusicTreePrefix(ITUNES_MUSIC, self.db.codecs.extensions('m4a'))
            self.register_prefix(itunes_prefix)
            self.load_user_config()

        def load_user_config(self):
            if not os.path.isfile(USER_PATH_CONFIG):
                return

            try:
                with open(USER_PATH_CONFIG, 'r') as config:

                    user_codecs = {}
                    for line in config:

                        try:
                            if line.strip() == '' or line[:1] == '#':
                                continue
                            (codec_name, paths) = [x.strip() for x in line.split('=', 1)]
                            paths = [x.strip() for x in paths.split(',')]
                        except ValueError:
                            self.log.debug('Error parsing line: %s' % line)
                            continue

                        user_codecs[codec_name] = paths

                    for codec_name in reversed(sorted(user_codecs.keys())):
                        paths = user_codecs[codec_name]

                        if codec_name == 'itunes':
                            codec = match_codec('m4a')
                        else:
                            codec = match_codec(codec_name)

                        if not codec:
                            continue

                        for path in reversed(paths):
                            prefix = MusicTreePrefix(path, self.db.codecs.extensions('aac'))

                            if codec_name == 'itunes':
                                self.register_prefix(prefix, prepend=False)
                            else:
                                self.register_prefix(prefix, prepend=True)

            except IOError, (ecode, emsg):
                raise PrefixError('Error reading %s: %s' % (USER_PATH_CONFIG, emsg))

        def index(self, prefix):
            if not isinstance(prefix, MusicTreePrefix):
                raise PrefixError('Prefix must be MusicTreePrefix instance')

            for index, existing in enumerate(self):
                if prefix.realpath == existing.realpath:
                    return index

            raise IndexError('Prefix is not registered')

        def register_prefix(self, prefix, extensions=[], prepend=False):
            if isinstance(prefix, basestring):
                prefix = MusicTreePrefix(prefix, extensions)

            if not isinstance(prefix, MusicTreePrefix):
                raise PrefixError('prefix must be string or MusicTreePrefix instance')

            try:
                index = self.index(prefix)
                if prepend and index != 0:
                    prefix = self.pop(index)
                    self.insert(0, prefix)

            except IndexError:
                if prepend:
                    self.insert(0, prefix)
                else:
                    self.append(prefix)

            return prefix

        def match_extension(self, extension, match_existing=False):
            for prefix in self:
                if match_existing and not os.path.isdir(prefix.path):
                    continue

                if prefix.match_extension(extension):
                    return prefix

            return None

        def match(self, path, match_existing=False):
            for prefix in self:

                if match_existing and not os.path.isdir(prefix.path):
                    continue

                if prefix.match(path):
                    return prefix

            return None

        def relative_path(self, path):
            prefix = self.match(path)
            if not prefix:
                return path

            return prefix.relative_path(path)

    def __getattr__(self, attr):
        return getattr(self.__instance, attr)

    def __setattr__(self, attr, value):
        return setattr(self.__instance, attr, value)




# coding=utf-8
"""Soundforest default settings

Default settings for configuration database and commands.

"""

import sys
import os

if sys.platform == 'darwin':
    SOUNDFOREST_USER_DIR = os.path.expanduser('~/Library/Application Support/Soundforest')
    SOUNDFOREST_CACHE_DIR = os.path.expanduser('~/Library/Caches/Soundforest')
else:
    SOUNDFOREST_USER_DIR = os.path.expanduser('~/.config/soundforest')
    SOUNDFOREST_CACHE_DIR = os.path.expanduser('~/.cache/soundforest')

DEFAULT_TREE_TYPES = {
  'songs': 'Full song audio files',
  'loops': 'Audio loops',
  'crates': 'Folders of exported audio files',
  'samples': 'Audio samples',
  'scratch': 'Scratch samples',
  'ringtones': 'Phone Ringtones',
  'notifications': 'Phone Notifications',
}

DEFAULT_CODECS = {

  'mp3': {
    'description': 'MPEG-1 or MPEG-2 Audio Layer III',
    'extensions':   ['mp3'],
    'encoders': [
      'lame --quiet -b 320 --vbr-new -ms --replaygain-accurate FILE OUTFILE',
    ],
    'decoders': [
      'lame --quiet --decode FILE OUTFILE',
    ],
  },

  'm4a': {
    'description': 'Advanced Audio Coding',
    'extensions': ['m4a', 'aac', 'mp4'],
    'encoders': [
      'afconvert -b 256000 --soundcheck-generate -f m4af -d aac FILE OUTFILE',
      'neroAacEnc -if FILE -of OUTFILE -br 256000 -2pass',
    ],
    'decoders': [
      'afconvert -f WAVE -d LEI16 FILE OUTFILE',
      'neroAacDec -if OUTFILE -of FILE',
      'faad -q -o OUTFILE FILE -b1',
    ],
    'testers': [
      'afconvert -f WAVE -d LEI16 FILE OUTFILE',
    ]
  },

  'm4r': {
    'description': 'AAC Alert/Ringtone for iOS',
    'extensions': ['m4r'],
    'encoders': [
      'afconvert -b 256000 --soundcheck-generate -f m4af -d aac FILE OUTFILE',
      'neroAacEnc -if FILE -of OUTFILE -br 256000 -2pass',
    ],
    'decoders': [
      'afconvert -f WAVE -d LEI16 FILE OUTFILE',
      'neroAacDec -if OUTFILE -of FILE',
      'faad -q -o OUTFILE FILE -b1',
    ],
  },

  'vorbis': {
    'description': 'Ogg Vorbis',
    'extensions': ['ogg', 'oga'],
    'encoders': [
      'oggenc --quiet -q 7 -o OUTFILE FILE',
    ],
    'decoders': [
      'oggdec --quiet -o OUTFILE FILE',
    ],
  },

  'alac': {
    'description': 'Apple Lossless Codec',
    'extensions': ['alac', 'm4a'],
    'encoders': [
      'afconvert -d alac FILE OUTFILE',
    ],
    'decoders': [
      'afconvert -f WAVE -d LEI16 FILE OUTFILE',
    ],
  },

  'flac': {
    'description': 'Free Lossless Audio Codec',
    'extensions': ['flac'],
    'encoders': [
      'flac -f --silent --verify --replay-gain --best -o OUTFILE FILE',
    ],
    'decoders': [
      'flac -f --silent --decode -o OUTFILE FILE',
    ],
    'testers': [
      'flac --silent --test FILE',
    ]
  },

  'wavpack': {
    'description': 'WavPack Lossless Audio Codec',
    'extensions': ['wv', 'wavpack'],
    'encoders': [ 'wavpack -yhx FILE -o OUTFILE' ],
    'decoders': [ 'wvunpack -yq FILE -o OUTFILE' ],
  },

  'caf': {
    'description': 'CoreAudio Format audio',
    'extensions':   ['caf'],
    'encoders': [
      'afconvert -f caff -d LEI16 FILE OUTFILE',
    ],
    'decoders': [
      'afconvert -f WAVE -d LEI16 FILE OUTFILE',
    ],
  },

  'aif': {
      'description': 'AIFF audio',
      'extensions':   ['aif', 'aiff'],
      'encoders': [
        'afconvert -f AIFF -d BEI16 FILE OUTFILE',
      ],
      'decoders': [
        'afconvert -f WAVE -d LEI16 FILE OUTFILE',
      ],
      },

  'wav': {
      'description': 'RIFF Wave Audio',
      'extensions':   ['wav'],
      'encoders': [
        'afconvert -f WAVE -d LEI16 FILE OUTFILE',
      ],
      'decoders': [
        'cp FILE OUTFILE',
      ],
  },

}



# coding=utf-8
"""Tree, album  and track

Abstraction of filesystem audio file trees, albums and tracks

"""

import os
import re
import shutil
import time

from soundforest import normalized, SoundforestError, TreeError
from soundforest.log import SoundforestLogger
from soundforest.formats import AudioFileFormat, path_string, match_codec, match_metadata
from soundforest.prefixes import TreePrefixes, PrefixError
from soundforest.metadata import CoverArt
from soundforest.tags import TagError
from soundforest.tags.albumart import AlbumArt, AlbumArtError
from soundforest.tags.tagparser import Tags


class IterableTrackFolder(object):
    """IterableTrackFolder model

    Abstract class for various iterable music items

    """

    def __init__(self, path, iterable):
        self.log = SoundforestLogger().default_stream
        self.__next = None
        self.__iterable = iterable
        if path in ['.', '']:
            path = os.path.realpath(path)
        self.path = path_string(path)
        self.prefixes = TreePrefixes()
        self.invalid_paths = []
        self.has_been_iterated = False
        setattr(self, iterable, [])

    def __getitem__(self, item):
        if not self.has_been_iterated:
            self.load()
        iterable = getattr(self, self.__iterable)
        return iterable[item]

    def __len__(self):
        iterable = getattr(self, self.__iterable)
        if len(iterable) == 0:
            self.load()
        if len(iterable) - len(self.invalid_paths) >= 0:
            return len(iterable) - len(self.invalid_paths)
        else:
            return 0

    def __iter__(self):
        return self

    def next(self):
        iterable = getattr(self, self.__iterable)
        if self.__next is None:
            self.__next = 0
            self.has_been_iterated = False
            if len(iterable) == 0:
                self.load()
        try:
            entry = iterable[self.__next]
            self.__next += 1
            path = os.path.join(entry[0], entry[1])
            try:
                return Track(path)
            except TreeError:
                if not self.invalid_paths.count(path):
                    self.invalid_paths.append(path)
                return self.next()
        except IndexError:
            self.__next = None
            self.has_been_iterated = True
            raise StopIteration

    def load(self):
        """Lazy loader of the iterable item"""
        iterable = getattr(self, self.__iterable)
        iterable.__delslice__(0, len(iterable))
        self.invalid_paths.__delslice__(0, len(self.invalid_paths))

    def relative_path(self, item=None):
        """Returns relative path of this iterable item"""

        if item is not None:
            if isinstance(item, Track):
                return self.prefixes.relative_path(item.path)
            else:
                return self.prefixes.relative_path(item)

        else:
            return self.prefixes.relative_path(self.path)

    def remove_empty_path(self, empty):
        """
        Remove empty directory and all empty parent directories
        """
        while True:
            if not os.path.isdir(empty):
                # Directory does not exist
                return
            if os.listdir(empty):
                # Directory is not empty
                return

            try:
                os.rmdir(empty)
            except OSError, (ecode, emsg):
                raise TreeError('Error removing empty directory %s: %s' % (empty, emsg))

            # Try to remove parent empty directory
            empty = os.path.dirname(empty)


class Tree(IterableTrackFolder):
    """Tree

    Audio file tree

    """

    def __init__(self, path):
        IterableTrackFolder.__init__(self, path, 'files')
        self.paths = {}
        self.empty_dirs = []
        self.relative_dirs = []

    def __len__(self):
        """
        Tree must be loaded to figure out it's length
        """
        if not self.has_been_iterated:
            self.has_been_iterated = True
            while True:
                try:
                    self.next()
                except StopIteration:
                    break
        return IterableTrackFolder.__len__(self)

    def __cmp_file_path__(self, a, b):
        if a[0] != b[0]:
            return cmp(a[0], b[0])
        return cmp(a[1], b[1])

    def load(self):
        """Load the albums and songs in the tree"""

        if not os.path.isdir(self.path):
            raise TreeError('Not a directory: %s' % self.path)

        self.log.debug('load tree: %s' % self.path)
        start = long(time.mktime(time.localtime()))

        IterableTrackFolder.load(self)
        self.paths = {}
        self.empty_dirs = []
        self.relative_dirs = []
        for (root, dirs, files) in os.walk(self.path, topdown=True):
            if files:
                self.files.extend((root, x) for x in files)
                for x in files:
                    self.paths[os.path.join(root, x)] = True
            elif not dirs:
                self.empty_dirs.append(root)

        self.relative_dirs = set(self.relative_path(x[0]) for x in self.files)
        self.files.sort(lambda x, y: self.__cmp_file_path__(x, y))

        stop = long(time.mktime(time.localtime()))
        self.log.debug('loaded %d files in %d seconds' % (len(self.files), (stop-start)))

    def filter_tracks(self, regexp=None, re_path=True, re_file=True, as_tracks=False):
        if not len(self.files):
            self.load()

        tracks = filter(lambda f: match_codec(f[1]), self.files)
        if regexp is not None:
            if not re_file and not re_path:
                raise TreeError('No matches if both re_file and re_path are False')
            if isinstance(regexp, basestring):
                regexp = re.compile(regexp)
            tracks = filter(lambda t:
                re_path and regexp.match(t[0]) or re_file and regexp.match(t[1]),
                tracks
            )

        if as_tracks:
            return [Track(os.path.join(t[0], t[1])) for t in tracks]
        else:
            return tracks

    @property
    def directories(self):
        return set(normalized(os.path.dirname(x)) for x in self.paths.keys())

    @property
    def realpaths(self):
        return dict((normalized(os.path.realpath(v)), True) for v in self.paths.keys())

    def contains(self, path):
        directory = os.path.dirname(path)
        filename = os.path.filename(path)

    def as_albums(self):
        if not self.has_been_iterated:
            self.load()
        return [Album(path) for path in sorted(set(d[0] for d in self.files))]

    def match(self, path):
        relative_path = self.relative_path(path)
        if not os.path.dirname(relative_path) in self.relative_dirs:
            return None

    def test(self, callback):
        for track in self:
            track.test(callback)


class Album(IterableTrackFolder):

    def __init__(self, path):
        IterableTrackFolder.__init__(self, path, 'files')
        self.metadata_files = []

    def __repr__(self):
        return 'album %s' % self.path

    def __getitem__(self, item):
        item = IterableTrackFolder.__getitem__(self, item)
        return Track(os.path.join(*item))

    def load(self):
        IterableTrackFolder.load(self)

        self.metadata_files = []
        for f in os.listdir(self.path):
            if match_codec(f) is not None:
                self.files.append((self.path, f))

            else:
                metadata = match_metadata(f)
                if metadata is not None:
                    self.metadata_files.append(MetaDataFile(os.path.join(self.path, f), metadata))

        self.files.sort()

    @property
    def mtime(self):
        return os.stat(self.path).st_mtime

    @property
    def ctime(self):
        return os.stat(self.path).st_ctime

    @property
    def atime(self):
        return os.stat(self.path).st_atime

    @property
    def metadata(self):
        if not self.has_been_iterated:
            self.load()
        return self.metadata_files

    @property
    def albumart(self):
        if not len(self.metadata):
            return None

        for m in self.metadata:
            if not hasattr(m, 'metadata'):
                raise TreeError('Invalid object types in self.metadata: %s' % m)
            if isinstance(m.metadata, CoverArt):
                return AlbumArt(m.path)

        return None

    def copy_metadata(self, target):
        if isinstance(target, basestring):
            target = Album(target)

        if not os.path.isdir(target.path):
            try:
                os.makedirs(target.path)
            except OSError, (ecode, emsg):
                raise TreeError('Error creating directory %s: %s' % (target.path, emsg))

        for m in self.metadata:
            dst_path = os.path.join(target.path, os.path.basename(m.path))

            if os.path.realpath(m.path) == os.path.realpath(dst_path):
                continue

            try:
                shutil.copyfile(m.path, dst_path)
            except OSError, (ecode, emsg):
                self.script.exit(1, 'Error writing %s: %s' % (dst_path, emsg))

        target.load()
        albumart = target.albumart
        if target.albumart:
            for track in target:
                tags = track.tags
                if tags is None:
                    continue

                if not tags.supports_albumart:
                    self.log.debug('no albumart support: %s' % track.path)
                    continue

                if tags.set_albumart(albumart):
                    self.log.debug('albumart: %s' % track)
                    tags.save()


class MetaDataFile(object):
    """MetaDataFile

    Metadata files, like album artwork, booklets and vendor specific
    analysis files.

    """

    def __init__(self, path, metadata=None):
        if metadata is None:
            metadata = match_metadata(path)
            if metadata is None:
                raise TreeError('Not a metadata file: %s' % path)

        self.path = path_string(path)

        self.extension = os.path.splitext(self.path)[1][1:].lower()
        if self.extension == '':
            self.extension = None

        self.metadata = metadata

    def __repr__(self):
        return '%s %s' % (
            self.metadata.description,
            self.filename
        )

    @property
    def filename(self):
        return os.path.basename(self.path)

class Track(AudioFileFormat):
    """Track

    Audio file track

    """

    def __init__(self, path):
        AudioFileFormat.__init__(self, path)
        self.prefixes = TreePrefixes()
        if self.codec is None:
            raise TreeError('Not a music file: %s' % self.path)
        self.tags_loaded = False
        self.file_tags = None

    @property
    def tags(self):
        if not self.tags_loaded:
            try:
                self.file_tags = Tags(self.path, fileformat=self)
                self.tags_loaded = True
            except TagError, emsg:
                raise TreeError('Error loading tags: %s' % emsg)

        return self.file_tags

    @property
    def relative_path(self):
        return self.prefixes.relative_path(os.path.realpath(self.path))

    @property
    def extension(self):
        return os.path.splitext(self.path)[1][1:]

    @property
    def album(self):
        return Album(os.path.dirname(self.path))

    @property
    def tracknumber_and_title(self):
        filename = os.path.splitext(os.path.basename(self.path))[0]
        try:
            tracknumber, title = filename.split(None, 1)
            tracknumber = int(tracknumber)
        except ValueError:
            tracknumber = None
            title = filename

        return tracknumber, title

    def get_album_tracks(self):
        path = os.path.dirname(self.path)
        extensions = CODECS[self.codec]['extensions']
        tracks = []

        for t in os.listdir(path):
            if os.path.splitext(t)[1][1:] not in extensions:
                continue
            tracks.append(Track(os.path.join(path, t)))

        return tracks

    def get_decoder_command(self, wav_path=None):
        if wav_path is None:
            wav_path = '%s.wav' % os.path.splitext(self.path)[0]
        if wav_path == self.path:
            raise TreeError('Trying to encode to itself')

        try:
            decoder = self.get_available_decoders()[0]
        except IndexError:
            raise TreeError('No available decoders for %s' % self.path)

        decoder = decoder.split()
        decoder[decoder.index('OUTFILE')] = wav_path
        decoder[decoder.index('FILE')] = self.path
        return decoder

    def get_encoder_command(self, wav_path=None):
        if wav_path is None:
            wav_path = '%s.wav' % os.path.splitext(self.path)[0]
        if wav_path == self.path:
            raise TreeError('Trying to encode to itself')

        try:
            encoder = self.get_available_encoders()[0]
        except IndexError:
            raise TreeError('No available encoders for %s' % self.path)

        encoder = encoder.split()
        encoder[encoder.index('OUTFILE')] = self.path
        encoder[encoder.index('FILE')] = wav_path
        return encoder

    def get_tester_command(self, tempfile_path):
        try:
            tester = self.get_available_testers()[0]
        except IndexError:
            raise TreeError('No available testers for %s' % self.path)

        tester = tester.split()
        tester[tester.index('FILE')] = self.path
        if tester.count('OUTFILE') == 1:
            tester[tester.index('OUTFILE')] = tempfile_path

        return tester

    def test(self, callback):
        try:
            cmd = self.get_tester_command(self.get_temporary_file(prefix='test', suffix='.wav'))
        except TreeError, emsg:
            callback(self, False, errors='No tester available for %s' % self.extension)
            return

        rv, stdout, stderr = self.execute(cmd)
        if rv == 0:
            callback(self, True, stdout=stdout, stderr=stderr)
        else:
            callback(self, False, stdout=stdout, stderr=stderr)



"""
XML schema representation of audio file metadata tags
"""

from lxml import etree as ET
from lxml.etree import XMLSyntaxError
from lxml.builder import E

from soundforest.tags.constants import parsedate

class XMLTagError(Exception):
    pass

XML_EXPORT_FIELDS = [
    'path',
    'album_artist',
    'sort_artist',
    'artist',
    'sort_album',
    'album',
    'sort_title',
    'title',
    'genre',
    'bpm',
    'year',
    'tracknumber',
    'comment',
    'composer',
    'copyright',
    'xid',
]

def XMLTrackNumberField(details):
    if details.has_key('totaltracks'):
        node = E('tracknumber',
            track=details['tracknumber'][0],
            total=details['totaltracks'][0]
        )
    else:
        node = E('tracknumber',
            track=details['tracknumber'][0],
        )
    return node

def XMLTrackYear(details):
    nodes = []
    for value in details['year']:
        value = parsedate(value)

        if value is None:
            continue
        nodes.append(E('year', '%d' % value.tm_year))
    return nodes

XML_FIELD_CLASSES = {
    'tracknumber': XMLTrackNumberField,
    'year': XMLTrackYear,
}

class XMLTags(dict):
    def __init__(self, data):
        self.tree = E('track')
        if not isinstance(data, dict):
            raise XMLTagError('Details must be dictionary')

        for k in XML_EXPORT_FIELDS:
            if k not in data.keys():
                continue

            if k in XML_FIELD_CLASSES.keys():
                nodes = XML_FIELD_CLASSES[k](data)
                if nodes is not None:
                    for node in nodes:
                        self.tree.append(node)

            else:
                for v in data[k]:
                    self.tree.append(E(k, v))

    def toxml(self):
        return ET.tostring(self.tree, pretty_print=True)


class XMLTrackTree(object):
    def __init__(self):
        self.tracks =  E('tracks')
        self.tree = E('soundforest', self.tracks)

    def append(self, xmltags):
        if not isinstance(xmltags, XMLTags):
            raise XMLTagError('xmltags must be XMLTags instance')
        self.tracks.append(xmltags.tree)

    def tostring(self):
        self.tracks.set('total', '%d' % len(self.tracks))
        return ET.tostring(self.tree, pretty_print=True)

